from xflask.type import Enum


class SysCode(Enum):

    SUCCESS          = '00000', 'Success'
    SYS_ERROR        = '10001', 'Internal Server Error'
    INVALID          = '10003', 'Request Invalid'
    NOT_FOUND        = '10004', 'Resource Not Found'
    EXISTED          = '10005', 'Resource Existed'
    AUTH_INVALID     = '10006', 'Authentication Invalid'



from flask_migrate import Migrate, MigrateCommand
from flask_script import Manager

from xflask.common.util import get_root_dir, import_modules
from xflask.sqlalchemy import db


class Migration(object):

    def __init__(self, server, model_pkgs=[], **args):
        self.server = server
        self.migrate = Migrate(server.app, db, directory='migration', **args)

        self.manager = Manager(server.app)
        self.manager.add_command('db', MigrateCommand)

        import_modules(get_root_dir(), model_pkgs)

    def run(self):
        self.manager.run()



import enum


class Enum(enum.Enum):

    def code(self):
        return self.value[0] if isinstance(self.value, tuple) else self.value

    def desc(self):
        return self.value[1] if isinstance(self.value, tuple) else None

    def type(self):
        return type(self.value[0])

    @classmethod
    def value_of(cls, code):
        result = None
        for e in cls:
            e_value = e.value if not isinstance(e.value, tuple) else e.value[0]
            if e_value == code:
                result = e
                break
        return result



import os
import sys
import importlib


def get_root_dir():
    return os.environ['PROJ_HOME']

def get_file(file):
    return os.path.join(get_root_dir(), file)

def import_modules(root_dir, model_pkgs):
    for model_pkg in model_pkgs:
        for module in os.listdir(root_dir + '/' + model_pkg.replace('.', '/')):
            if module == '__init__.py' or module[-3:] != '.py':
                continue

            model_namespace = model_pkg + '.' + module[:-3]

            importlib.import_module(model_namespace)

def setup_root_dir(root_dir_name):
    cur_dir = os.getcwd()
    while True:
        if os.path.basename(cur_dir) == root_dir_name:
            root_dir = cur_dir
            break
        else:
            cur_dir = os.path.dirname(cur_dir)

    os.environ['PROJ_HOME'] = root_dir
    sys.path.append(root_dir)



import enum
from datetime import datetime

from xflask.common.util.date_util import to_date_str
from xflask.type import Enum as Enum


def get_attr(obj, attr_name):
    return obj[attr_name] if isinstance(obj, dict) else getattr(obj, attr_name)

def to_dict(obj):
    if not hasattr(obj, '__dict__') or isinstance(obj, enum.Enum):
        return obj

    result = {}
    for key, val in obj.__dict__.items():
        if key.startswith("_"):
            continue

        element = []
        if isinstance(val, list):
            for item in val:
                element.append(to_dict(item))
        else:
            element = to_dict(val)

        result[key] = element

    return result

def serialize(obj, show=[], hidden=[], dept=0):
    if not hasattr(obj, '__dict__'):
        # primitive type
        if isinstance(obj, Enum):
            return obj.code()
        elif isinstance(obj, enum.Enum):
            return obj.value
        elif isinstance(obj, datetime):
            return to_date_str(obj)
        else:
            return obj

    if hasattr(obj, '_hidden'):
        _hidden = getattr(obj, '_hidden')
        hidden.extend(_hidden)
    if hasattr(obj, '_show'):
        _show = getattr(obj, '_show')
        show.extend(_show)

    hidden = [e for e in hidden if e not in show]

    if hasattr(obj, 'serialize'):
        return obj.serialize(show, hidden)

    result = {}
    for key, val in obj.__dict__.items():
        if key.startswith("_") or key in hidden:
            continue

        element = []
        if isinstance(val, list):
            item_show = [e.split('.')[1] for e in show if e.split('.')[0] == key] if dept > 0 else show
            item_hidden = [e.split('.')[1] for e in hidden if e.split('.')[0] == key] if dept > 0 else hidden

            for item in val:
                element.append(serialize(item, item_show, item_hidden, dept + 1))

        elif hasattr(val, 'serialize'):
            item_show = [e.split('.')[1] for e in show if e.split('.')[0] == key] if dept > 0 else show
            item_hidden = [e.split('.')[1] for e in hidden if e.split('.')[0] == key] if dept > 0 else hidden

            element = serialize(val, item_show, item_hidden, dept + 1)
        else:
            element = serialize(val, dept=dept + 1)

        result[key] = element

    return result



import logging
from flask import render_template
from flask import request
from marshmallow.exceptions import ValidationError
from werkzeug.exceptions import NotFound, BadRequest, MethodNotAllowed

from xflask.exception import Exception as Sys_Exception
from xflask.type.sys_code import SysCode
from xflask.web.response import Response


class ErrorHandler(object):

    def init(self, server):
        pass

    def handler_404(self, e):
        pass

    def handler_400(self, e):
        pass

    def handler_500(self, e):
        pass


class SimpleErrorHandler(ErrorHandler):
    DEF_API_ROUTE = '/api'
    DEF_TEMPLATE_FOLDER = ''

    def __init__(self, api_route=DEF_API_ROUTE, template_folder=DEF_TEMPLATE_FOLDER):
        self.api_route = api_route
        self.template_folder = template_folder

    def init(self, server):
        self.logger = logging.getLogger(self.__class__.__name__)

        server.app.register_error_handler(NotFound, self.handler_404)
        server.app.register_error_handler(BadRequest, self.handler_400)
        server.app.register_error_handler(MethodNotAllowed, self.handler_400)
        server.app.register_error_handler(ValidationError, self.handler_400)
        server.app.register_error_handler(Exception, self.handler_500)

    def handler_404(self, e):
        self.logger.exception('404 error')

        if request.path.startswith(self.api_route):
            return Response.fail(SysCode.NOT_FOUND).to_dict()
        else:
            return render_template(self.template_folder + '404.html')

    def handler_400(self, e):
        self.logger.exception('400 error')

        if request.path.startswith(self.api_route):
            if isinstance(e, ValidationError):
                return Response.fail(SysCode.INVALID, e.messages).to_dict()
            else:
                return Response.fail(SysCode.INVALID).to_dict()
        else:
            return render_template(self.template_folder + '400.html')

    def handler_500(self, e):
        self.logger.exception('500 error')

        if request.path.startswith(self.api_route):
            code = e.code if isinstance(e, Sys_Exception) else SysCode.SYS_ERROR
            return Response.fail(code).to_dict()
        else:
            return render_template(self.template_folder + '500.html')




import sqlalchemy.types as types

from xflask.type import Enum


class StringEnum(types.TypeDecorator):
    impl = types.String

    def __init__(self, enum_type: Enum, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.enum_type = enum_type

    def process_bind_param(self, value, dialect):
        return None if value is None else value.code()

    def process_result_value(self, value, dialect):
        return self.enum_type.value_of(value)


class IntegerEnum(StringEnum):
    impl = types.Integer



import os
import re

import yaml


class Configuration:

    def __init__(self, conf_file):
        with open(conf_file, 'r') as file:
            param_matcher = re.compile(r'.*\$\{([^}^{]+)\}.*')

            def param_constructor(loader, node):
                value = node.value

                params = param_matcher.findall(value)
                for param in params:
                    try:
                        param_value = os.environ[param]
                        return value.replace('${' + param + '}', param_value)
                    except Exception:
                        pass

                return value

            class VariableLoader(yaml.SafeLoader):
                pass

            VariableLoader.add_implicit_resolver('!param', param_matcher, None)
            VariableLoader.add_constructor('!param', param_constructor)

            self.cfg = yaml.load(file, Loader=VariableLoader)

    def exist(self, key):
        return True if self.get(key) is not None else False

    def get(self, key, default=None):
        value = None
        try:
            keys = key.split(':')
            if len(keys) > 1:
                value = self.__get_nest_value(self.cfg[keys[0]], keys[1:])
            else:
                value = self.cfg[key]

            if value is None and default is not None:
                value = default

        except Exception:
            pass

        return value

    def set(self, key, value):
        self.cfg.__setitem__(key, value)

    def __get_nest_value(self, map_, keys):
        if len(keys) > 1:
            return self.__get_nest_value(map_[keys[0]], keys[1:])
        else:
            return map_[keys[0]]

    def dump(self):
        return yaml.dump(self.cfg, Dumper=yaml.Dumper)



import enum
from datetime import datetime

from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.sql import func

from xflask.common.util.date_util import to_date_str
from xflask.sqlalchemy import db
from xflask.type.enum import Enum
from xflask.web.security import get_current_user


class Model(db.Model):
    __abstract__ = True

    def serialize(self, show=[], hide=[], dept=0):
        return self.to_dict(show, hide, dept, True)

    def to_dict(self, show=[], hide=[], dept=0, serialize=False):
        hidden = self._hidden_fields if hasattr(self, "_hidden_fields") else []
        hidden.extend([e for e in hide if '.' not in e])
        hidden = [e for e in hidden if e not in show]

        ret_data = {}

        # fields
        columns = self.__table__.columns.keys()
        for key in columns:
            if key.startswith("_") or key in hidden:
                continue

            value = getattr(self, key)
            if serialize is True:
                value = self._custom_serialization(value)

            ret_data[key] = value

        # relationships
        if dept > 0:
            relationships = self.__mapper__.relationships.keys()
            for key in relationships:
                if key.startswith("_") or key in hidden:
                    continue

                _show = [e.split('.')[1] for e in show if e.split('.')[0] == key]
                _hide = [e.split('.')[1] for e in hide if e.split('.')[0] == key]

                is_list = self.__mapper__.relationships[key].uselist
                if is_list:
                    items = getattr(self, key)
                    if self.__mapper__.relationships[key].query_class is not None:
                        if hasattr(items, "all"):
                            items = items.all()

                    ret_data[key] = []

                    for item in items:
                        ret_data[key].append(
                            item.to_dict(
                                show=list(_show),
                                hide=list(_hide),
                                dept=dept - 1
                            )
                        )
                else:
                    if (
                            self.__mapper__.relationships[key].query_class is not None
                            or self.__mapper__.relationships[key].instrument_class
                            is not None
                    ):
                        item = getattr(self, key)
                        if item is not None:
                            ret_data[key] = item.to_dict(
                                show=list(_show),
                                hide=list(_hide),
                                dept=dept - 1
                            )
                        else:
                            ret_data[key] = None
                    else:
                        value = getattr(self, key)
                        if serialize is True:
                            value = self._custom_serialization(value)

                    ret_data[key] = value

        return ret_data

    def _custom_serialization(self, obj):
        if isinstance(obj, Enum):
            return obj.code()
        elif isinstance(obj, enum.Enum):
            return obj.value
        elif isinstance(obj, datetime):
            return to_date_str(obj)
        else:
            return obj


class AuditModel(Model):
    __abstract__ = True

    created_at = db.Column(db.DateTime)
    modified_at = db.Column(db.DateTime)

    @declared_attr
    def created_by(cls):
        return db.Column(db.Integer)

    @declared_attr
    def modified_by(cls):
        return db.Column(db.Integer)

    @staticmethod
    def before_insert(mapper, connection, instance):
        user = get_current_user() or {}
        instance.created_at = func.now()
        instance.created_by = user.get('id')

    @staticmethod
    def before_update(mapper, connection, instance):
        user = get_current_user() or {}
        instance.modified_at = func.now()
        instance.modified_by = user.get('id')

    @classmethod
    def __declare_last__(cls):
        db.event.listen(cls, 'before_insert', cls.before_insert)
        db.event.listen(cls, 'before_update', cls.before_insert)


class SoftModel:
    deleted_at = db.Column(db.DateTime)

    @declared_attr
    def deleted_by(cls):
        return db.Column(db.Integer)

    @staticmethod
    def before_delete(mapper, connection, instance):
        user = get_current_user() or {}
        instance.deleted_at = func.now()
        instance.deleted_by = user.get('id')

    @classmethod
    def __declare_last__(cls):
        db.event.listen(cls, 'before_delete', cls.before_delete)



import collections
import logging.config
import os
import re

import yaml


class Logger(object):

    def __init__(self, log_file):
        with open(log_file, 'r') as f:
            param_matcher = re.compile(r'.*\$\{([^}^{]+)\}.*')

            def param_constructor(loader, node):
                value = node.value

                params = param_matcher.findall(value)
                for param in params:
                    try:
                        param_value = os.environ[param]
                        return value.replace('${' + param + '}', param_value)
                    except Exception:
                        pass

                return value

            class VariableLoader(yaml.SafeLoader):
                pass

            VariableLoader.add_implicit_resolver('!param', param_matcher, None)
            VariableLoader.add_constructor('!param', param_constructor)

            config = yaml.load(f.read(), Loader=VariableLoader)

            self.init_log_dir(config)

            logging.config.dictConfig(config)

    def init_log_dir(self, dict_):
        for k, v in dict_.items():
            if k == 'filename':
                filename = v
                file_dir = os.path.dirname(os.path.abspath(filename))

                if not os.path.exists(file_dir):
                    os.makedirs(file_dir)
            elif isinstance(v, collections.Mapping):
                self.init_log_dir(v)



class Param(object):
    def __init__(self, name=None, required=True):
        self.name = name
        self.required = required


class Header(object):

    def __init__(self, name=None, required=True):
        self.name = name
        self.required = required


## JSON ##

class JsonBody(object):

    def __init__(self, type=None, exclude=[]):
        self.type = type
        self.exclude = exclude


class JsonField(object):

    def __init__(self, name, required=True):
        self.name = name
        self.required = required


## FORM ##

class FormBody(object):

    def __init__(self, type=None, exclude=[]):
        self.type = type
        self.exclude = exclude


class FormField(object):

    def __init__(self, name, required=True):
        self.name = name
        self.required = required





from xflask.common.util import to_dict
from xflask.type import Enum
from xflask.type.sys_code import SysCode


class Response(object):

    def __init__(self, code: Enum, data=None, show=[], hidden=[]):
        self.status = code.code()
        self.message = code.desc()
        self.data = data

        # for serialization
        self._show = show
        self._hidden = hidden

    def to_dict(self):
        return to_dict(self)

    @classmethod
    def success(cls, data=None, show=[], hidden=[]):
        return Response(SysCode.SUCCESS, data, show, hidden)

    @classmethod
    def fail(cls, code=None, data=None, show=[], hidden=[]):
        if code is not None:
            return Response(code, data, show, hidden)
        else:
            return Response(SysCode.SYS_ERROR, data, show, hidden)

    @classmethod
    def not_found(cls):
        return Response(SysCode.NOT_FOUND)

    @classmethod
    def existed(cls):
        return Response(SysCode.EXISTED)

    @classmethod
    def invalid(cls):
        return Response(SysCode.INVALID)



from xflask.common.util.obj_util import to_dict
from xflask.component import Component
from xflask.sqlalchemy import db, transactional


class Dao(Component):

    def __init__(self, model):
        self.model = model

    def exist(self, id):
        return self.query().filter_by(id=id).scalar() is not None

    def get(self, id):
        return self.query().get(id)

    def get_all(self):
        return self.query().all()

    def query(self, **models):
        if models is None or len(models) == 0:
            return db.session.query(self.model)
        else:
            return db.session.query(**models)

    @transactional()
    def insert(self, obj):
        if isinstance(obj, dict):
            obj = self.model(**obj)

        db.session.add(obj)

    @transactional()
    def update(self, obj):
        if isinstance(obj, dict):
            self.query().filter_by(id=obj['id']).update(obj)
        else:
            self.query().filter_by(id=obj.id).update(to_dict(obj))

    @transactional()
    def delete(self, obj):
        if isinstance(obj, (int, float)):
            self.query().filter_by(id=obj).delete()
        else:
            db.session.delete(obj)

    def begin(self, subtransactions=True, nested=False):
        db.session.begin(subtransactions=subtransactions, nested=nested)

    def begin_nested(self):
        db.session.begin_nested()

    def flush(self, objs):
        db.session.flush(objs)

    def merge(self, obj):
        return db.session.merge(obj)

    def commit(self):
        db.session.commit()

    def rollback(self):
        db.session.rollback()



from marshmallow import Schema, EXCLUDE


class Vo(object):

    def __init__(self, *args, **kwargs):
        if not hasattr(self.__class__, '__annotations__'):
            return

        if args is not None and len(args) > 0:
            for idx, name in enumerate(self.__annotations__.keys()):
                value = None if idx >= len(args) else args[idx]
                self.__dict__[name] = value
        else:
            for name in self.__annotations__.keys():
                value = kwargs.get(name)
                self.__dict__[name] = value

    @classmethod
    def _get_schema(cls, exclude=[]):
        if not hasattr(cls, '__annotations__'):
            raise Exception(cls.__name__ + ' has no annotations')

        annotations_map = {}

        annotations = cls.__annotations__
        for name, annotation in annotations.items():
            if name in exclude:
                continue
            annotations_map[name] = annotation

        if len(exclude) > 0:
            return Schema.from_dict(annotations_map)(unknown=EXCLUDE)
        else:
            return Schema.from_dict(annotations_map)()

    @classmethod
    def validate(cls, obj: dict, exclude=[]):
        schema = cls._get_schema(exclude)
        return schema.validate(obj)

    @classmethod
    def serialize(cls, obj, exclude=[]):
        schema = cls._get_schema(exclude)
        return schema.dump(obj)

    def serialize_(self, exclude=[]):
        schema = self.__class__._get_schema(exclude)
        return schema.dump(self)

    @classmethod
    def deserialize(cls, obj: dict, exclude=[]):
        schema = cls._get_schema(exclude)
        return cls(**schema.load(obj))

    def deserialize_(self, obj: dict, exclude=[]):
        schema = self.__class__._get_schema(exclude)
        return self.__class__(**schema.load(obj))

    @classmethod
    def deserialize_as_dict(cls, obj: dict, exclude=[]):
        schema = cls._get_schema(exclude)
        return schema.load(obj)




from xflask.common import *

def get_current_user():
    auth_manager = get_xflask().auth_manager
    return auth_manager.get_current_user()

def logout():
    auth_manager = get_xflask().auth_manager
    return auth_manager.logout()



from xflask.marshmallow import Field
from xflask.type import Enum


class Enum(Field):

    def __init__(self, enum_type: Enum, **kwargs):
        super().__init__(**kwargs)
        self.enum_type = enum_type

    def _serialize(self, value, attr, obj, **kwargs):
        return value.code()

    def _deserialize(self, value, attr, data, **kwargs):
        super()._deserialize(value, attr, data, **kwargs)
        return self.enum_type.value_of(value)



from xflask.component import Component
from xflask.dao import Dao


class Service(Component):
    ...


class CrudService(Service):

    def __init__(self, dao: Dao):
        self.dao = dao

    def exist(self, id):
        return self.dao.exist(id)

    def get(self, id):
        return self.dao.get(id)

    def get_all(self):
        return self.dao.get_all()

    def query(self, models):
        return self.dao.query(models)

    def create(self, obj):
        self.dao.insert(obj)

    def update(self, obj):
        self.dao.update(obj)

    def delete(self, obj):
        self.dao.delete(obj)

    def begin(self, subtransactions=True, nested=False):
        self.dao.begin(subtransactions=subtransactions, nested=nested)

    def begin_nested(self):
        self.dao.begin_nested()

    def flush(self, objs):
        self.dao.flush(objs)

    def commit(self):
        self.dao.commit()

    def rollback(self):
        self.dao.rollback()
